/* This file forks two processes. A TCP server, and multiple clients
   that will use the TCP server. */
#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>
#include <string.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <netdb.h>
#include <unistd.h>
#include <fcntl.h>

#include "hcu_headers.h"
HCU_STATIC_UPDATE_FILE("./libtest_network_socket_v1.c_v1_to_v2.hcupatch.so.2",2);


#define SERVER_HOSTNAME "127.0.0.1"
#define SERVER_PORT 42166
#define SERVER_SOCKET_BACKLOG 50
#define SERVER_SOCKET_BUFFER_SIZE 1024
#define NUMBER_OF_CLIENTS 10
#define CLIENT_SOCKET_BUFFER_SIZE SERVER_SOCKET_BUFFER_SIZE

#define SERVER_COMMAND_HELLO "Hello"
#define SERVER_COMMAND_EXIT "Exit"
#define SERVER_REPLY_OK "HCU OK"
#define SERVER_REPLY_NOTOK "HCU NOT ok"
#define SERVER_REPLY_IWILLEXIT "HCU I will exit"


void start_the_clients();
void *client_thread(void *arg);
void client_send_data(int client_number, int sockfd);

void start_the_server();
void server_sigchld_handler(int s);
void *server_thread(void *arg);
void server_process_data(char *incoming_buffer, int sockfd);



int main()
{
  pid_t pid;

  pid = fork();
  if ( pid < 0 ) {
    printf("fork error: process cannot be forked\n");
    exit(1);
  } else if ( pid == 0 ) {
    // Child
    /* Give the parent some time to start the server */
    sleep(2);
    start_the_clients();
  } else{
    // Parent process
    start_the_server();
  }
}


/* ===========================================================================
   Client implementation
   ===========================================================================
*/

void start_the_clients()
{
  int number_of_clients = NUMBER_OF_CLIENTS;
  int counter = 0;

  /* Start all the clients */
  while (counter < NUMBER_OF_CLIENTS) {
    pthread_t client_thread_id;
    pthread_create(&client_thread_id, NULL, client_thread, (void *) counter);
    counter++;
  }

  /* Wait indefinitely */
  while (1) {
    /* Sleep 0.1 seconds */
    usleep(100000);
  }
  
}


void *client_thread(void *arg)
{
  int sockfd;
  struct hostent *he;
  struct sockaddr_in their_addr; // connector's address information 


  /* 
   * Setup a socket
   */
  if ((he=gethostbyname(SERVER_HOSTNAME)) == NULL) {  // get the host info 
    perror("client: gethostbyname error\n");
  }

  if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
    perror("client: socket error");
  }

  their_addr.sin_family = AF_INET;    // host byte order 
  their_addr.sin_port = htons(SERVER_PORT);  // short, network byte order 
  their_addr.sin_addr = *((struct in_addr *)he->h_addr);
  memset(&(their_addr.sin_zero), '\0', 8);  // zero the rest of the struct 

  if (connect(sockfd, (struct sockaddr *)&their_addr,
	      sizeof(struct sockaddr)) == -1) {
    perror("client: connecting to server failed");
    return;
  }

  /*
   * Send/receive data to the server
   */
  while(1) {
    client_send_data((int) arg, sockfd);
    /* Sleep 0.1 seconds */
    usleep(100000);
  }

  close(sockfd);
}



void client_send_data(int client_number, int sockfd)
{
  int number_of_bytes_received;
  char incoming_buffer[CLIENT_SOCKET_BUFFER_SIZE];

  memset(incoming_buffer, 0, CLIENT_SOCKET_BUFFER_SIZE);

  /* Send something to the server. If we are the last client, check if
     the server supports a new command. */
  if (client_number == NUMBER_OF_CLIENTS - 1) {
    if (send(sockfd, SERVER_COMMAND_EXIT, strlen(SERVER_COMMAND_EXIT), 0) == -1)
      perror("client: client_send_data SERVER_COMMAND_EXIT error");
  } else {
    if (send(sockfd, SERVER_COMMAND_HELLO, strlen(SERVER_COMMAND_HELLO), 0) == -1)
      perror("client: client_send_data SERVER_COMMAND_HELLO error");
  }

 try_receiving_again:
  number_of_bytes_received = recv(sockfd, incoming_buffer, CLIENT_SOCKET_BUFFER_SIZE, 0);
  if (number_of_bytes_received == -1) {
    perror("client: hcu receive error");
    /* Sleep 0.2 seconds, and try again */
    usleep(200000);
    printf("client: will try to receive again hcu\n");
    goto try_receiving_again;
  }
  incoming_buffer[number_of_bytes_received] = '\0';

  // Report the response of the server
  printf("client: hcu response from server is: '%s'\n", incoming_buffer);

  if (strcmp(incoming_buffer, SERVER_REPLY_IWILLEXIT) == 0) {
    // The server has been updated. It reported that it will exit, and
    // so should we.
    exit(0);
  }
}

/* ===========================================================================
   Server implementation
   ===========================================================================
*/
void start_the_server()
{
  int sockfd, new_fd;  // listen on sock_fd, new connection on new_fd
  struct sockaddr_in my_addr;    // my address information
  struct sockaddr_in their_addr; // connector's address information
  int sin_size;
  struct sigaction sa;
  int yes=1;
  pthread_t  tid;             /* variable to hold thread ID */

  if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
    perror("server: socket error");
    exit(1);
  }

  my_addr.sin_family = AF_INET;         // host byte order
  my_addr.sin_port = htons(SERVER_PORT);     // short, network byte order
  my_addr.sin_addr.s_addr = INADDR_ANY; // automatically fill with my IP
  memset(&(my_addr.sin_zero), '\0', 8); // zero the rest of the struct
  
  while (bind(sockfd, (struct sockaddr *)&my_addr, sizeof(struct sockaddr)) == -1) {
    if (errno == EADDRINUSE) {
      printf("server: error binding on port %d. Will try again hcu\n", SERVER_PORT);
      usleep(10000);
    } else {
      perror("server: bind error");
      exit(1);
    }
  }

  if (listen(sockfd, SERVER_SOCKET_BACKLOG) == -1) {
    perror("server: listen error");
    exit(1);
  }

  sa.sa_handler = server_sigchld_handler; // reap all dead processes
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = SA_RESTART;
  if (sigaction(SIGCHLD, &sa, NULL) == -1) {
    perror("server: sigaction error");
    exit(1);
  }

  while(1) {  // main accept() loop
    sin_size = sizeof(struct sockaddr_in);
    if ((new_fd = accept(sockfd, (struct sockaddr *)&their_addr,
			 &sin_size)) == -1) {
      /* Sleep 0.01 seconds */
      usleep(10000);
      continue;
    }
    
    //printf("server: got connection from %s\n", inet_ntoa(their_addr.sin_addr));

    pthread_create(&tid, NULL, server_thread, (void *)new_fd);

  }

  close(sockfd);
}


void server_sigchld_handler(int s)
{
  while(wait(NULL) > 0);
}



void *server_thread(void *sockfd)
{
  int number_of_bytes_received;
  char incoming_buffer[SERVER_SOCKET_BUFFER_SIZE];
  int number_of_server_commands_handled = 0;

  memset(incoming_buffer, 0, SERVER_SOCKET_BUFFER_SIZE);

  /* Continuously process data from the client */
  while (1) {
    /* Receive data from a client */
    printf("server: Thread %d is WAITING to receive hcu\n", pthread_self());
  try_receiving_again:
    number_of_bytes_received = recv((int) sockfd, incoming_buffer, SERVER_SOCKET_BUFFER_SIZE, 0);
    if ( number_of_bytes_received == -1) {
      perror("server: hcu receive error");
      /* Sleep 0.2 seconds, and try again */
      usleep(200000);
      printf("server: will try to receive again hcu\n");
      goto try_receiving_again;
    }
    incoming_buffer[number_of_bytes_received] = '\0';
    printf("server: Thread %d RECEIVED %d bytes hcu for '%s'\n", pthread_self(), number_of_bytes_received, incoming_buffer);
    server_process_data(incoming_buffer, (int) sockfd);
    number_of_server_commands_handled++;

    /* Update the server after a while. */
    if (number_of_server_commands_handled == 3) {
      printf("server: number_of_server_commands_handled: %d HCU must update now\n", number_of_server_commands_handled);
      HCU_STATIC_REQUEST_UPDATE_IMMEDIATE();
    }
  }
}


void server_process_data(char *incoming_buffer, int sockfd)
{
  if (strcmp(incoming_buffer, SERVER_COMMAND_HELLO) == 0) {

    if (send(sockfd, SERVER_REPLY_OK, strlen(SERVER_REPLY_OK), 0) == -1)
      perror("server: server_process_data SERVER_COMMAND_HELLO error");

  } else {
    // An invalid command was received.
    if (send(sockfd, SERVER_REPLY_NOTOK, strlen(SERVER_REPLY_NOTOK), 0) == -1)
      perror("server: server_process_data SERVER_REPLY_NOTOK error");
  }
}
