# UpStare is a system that can apply immediate dynamic software updates
# in multi-threaded userspace applications using stack
# reconstruction. It is Copyright (C) 2006 by Kristis Makris
# <kristis.makris@asu.edu,mkgnu@mkgnu.net>.

# In binary form produced and digitally signed specifically by Kristis
# Makris, it is freely distributed for personal and academic/research
# use. It is NOT distributed for commercial use. No other binary forms
# can be distributed.

# The source code and documentation, either in electronic or hardcopy
# format, may not be used for any purpose whatsoever without the written
# permission of Kristis Makris.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#
# This is a shell library containing common testing logic
#
DEBUG=0


# If debugging the tests is enabled, do not start an update thread in
# the programs. This will make debugging them in gdb easier, since the
# program will be single threaded. Well, at least for programs that
# were single-threaded to begin with
BINARY_FULL="/usr/bin/hcucc.pl"
BINARY_SIMPLE="/usr/bin/hcucil.pl"
BINARY_CONFIGURATION="/usr/bin/hcubasic.pl --doConfiguration"
BINARY_STACK_RECONSTRUCTION="/usr/bin/hcu_cilly.sh --noPrintLn --doCommonUpdate --doStackReconstruction --save-temps -I /usr/include/upstare/runtime -lhcu -ldl -lpthread"
BINARY_WRAP_CALLS="/usr/bin/hcu_cilly.sh --noPrintLn --doWrapCalls --save-temps -I /usr/include/upstare/runtime -lhcu -ldl -lpthread"
BINARY_DYNAMIC_SAFETY="/usr/bin/hcubasic.pl --doCommonUpdate --doWrapCalls --doDynamicSafety"
BINARY_UPDATE_POINTS="/usr/bin/hcu_cilly.sh --noPrintLn --doInsertUpdatePoints --save-temps -I /usr/include/upstare/runtime -lhcu -ldl -lpthread"
BINARY_CONVERT_BLOCKING="/usr/bin/hcu_cilly.sh --noPrintLn --doBlockingSystemCallConversion --save-temps -I /usr/include/upstare/runtime -lhcu -ldl -lpthread"
BINARY_APPLY="/usr/bin/hcuapply"
BINARY_DIFF="/usr/bin/hcudiff.sh"
BINARY_BUILD_PATCH="/usr/bin/hcu_build_patch.sh"
BINARY_CC=gcc


if [ $DEBUG = 1 ]; then
    BINARY_FULL="$BINARY_FULL --no-update-thread"
fi


test_diff()
{
    SOURCE_FILE=$1
    FLAGS=$2

    CIL_FILE=`echo -n $SOURCE_FILE| sed -e 's/.c$/.cil.c/'`

    # Preprocess
    $BINARY_FULL "$FLAGS" $SOURCE_FILE
    mv $CIL_FILE $SOURCE_FILE.output

    cat $SOURCE_FILE.output

    # If the output does not match the expected, the test failed
    diff -u $SOURCE_FILE.expected $SOURCE_FILE.output || test_diff_failed
    echo PASSED
}



test_configuration()
{
    SOURCE_FILE=$1
    CONFIGURATION_FILE=$2
    FLAGS=$3

    CIL_FILE=`echo -n $SOURCE_FILE| sed -e 's/.c$/.c.hcucc.cil.c/'`

    # Preprocess
    $BINARY_CONFIGURATION --compiler-configuration-file=$CONFIGURATION_FILE "$FLAGS" $SOURCE_FILE
    mv $CIL_FILE $SOURCE_FILE.output

    cat $SOURCE_FILE.output

    # If the output does not match the expected, the test failed
    diff -u $SOURCE_FILE.expected $SOURCE_FILE.output || test_diff_failed
    echo PASSED
}



test_compiler_wrapper()
{
    TEST_TYPE=$1
    SOURCE_FILE=$2
    ARGUMENTS=$3

    rm $SOURCE_FILE.c.hcucc.i
    rm a.out
    echo $BINARY_FULL $ARGUMENTS
    $BINARY_FULL --base-version $ARGUMENTS
    RETVAL=$?

    if [ "${TEST_TYPE}" = "wrap_source" ]; then
	#
	# We are testing if the source was wrapped correctly
	#
	if [ $RETVAL = 0 ]; then
	# The compiler wrapper picked up the file and correctly produced output
	    echo PASSED
	else
	    echo FAILED
	fi
    elif [ "${TEST_TYPE}" = "compile_source" ]; then
	#
	# We are testing if the source was compiled correctly
	#
	if [ -e a.out ]; then
	# The compiler wrapper worked and the source code was compiled
	    echo PASSED
	else
	    echo FAILED
	fi
    fi    
}



test_compiler_compliance()
{
    SOURCE_FILE=$1
    FLAGS="$2"

    # First compile the program with a proper compiler
    rm a.out
    $BINARY_CC $FLAGS -g $SOURCE_FILE > $SOURCE_FILE.cc.compile.output 2>&1
    # Run the program 
    ./a.out > $SOURCE_FILE.cc.execute.output 2>&1

    # Now compile the program with the CIL compiler
    rm a.out
    $BINARY_SIMPLE --notrace --nostats $FLAGS -g $SOURCE_FILE > $SOURCE_FILE.cil.compile.output 2>&1
    # Run the program 
    ./a.out > $SOURCE_FILE.cil.execute.output 2>&1


    # Compare the output from the compilation. There should be no
    # additional warnings or errors.
    diff -u $SOURCE_FILE.cc.compile.output $SOURCE_FILE.cil.compile.output || test_diff_failed

    # Compare the output from the execution.
    diff -u $SOURCE_FILE.cc.execute.output $SOURCE_FILE.cil.execute.output || test_diff_failed

    echo PASSED
}



test_build()
{
    DIRECTORY_NAME=$1

    cd $DIRECTORY_NAME
    make clean
    make
    cd ..

    #
    # We are testing if the source in this directory was compiled and
    # linked correctly
    #
    if [ -e $DIRECTORY_NAME/$DIRECTORY_NAME ]; then
	# The compiler wrapper worked and the source code was compiled
	echo PASSED
    else
	echo FAILED
    fi
}



test_wrap_calls()
{
    SOURCE_FILE=$1
    FLAGS=$2

    CIL_FILE=`echo -n $SOURCE_FILE| sed -e 's/.c$/.cil.c/' `

    # Preprocess
    $BINARY_WRAP_CALLS "$FLAGS" $SOURCE_FILE
    mv $CIL_FILE $SOURCE_FILE.output

    cat $SOURCE_FILE.output

    # If the output does not match the expected, the test failed
    diff -u $SOURCE_FILE.expected $SOURCE_FILE.output || test_diff_failed
    echo PASSED
}



test_dynamic_safety()
{
    SOURCE_FILE=$1
    FLAGS=$2

    CIL_FILE=`echo -n $SOURCE_FILE| sed -e 's/.c$/.c.hcucc.cil.c/' `

    # Preprocess
    $BINARY_DYNAMIC_SAFETY "$FLAGS" $SOURCE_FILE
    mv $CIL_FILE $SOURCE_FILE.output

    cat $SOURCE_FILE.output

    # If the output does not match the expected, the test failed
    diff -u $SOURCE_FILE.expected $SOURCE_FILE.output || test_diff_failed
    echo PASSED
}



test_update_points()
{
    SOURCE_FILE=$1
    FLAGS=$2

    CIL_FILE=`echo -n $SOURCE_FILE| sed -e 's/.c$/.cil.c/' `

    # Preprocess
    $BINARY_UPDATE_POINTS "$FLAGS" $SOURCE_FILE
    mv $CIL_FILE $SOURCE_FILE.output

    cat $SOURCE_FILE.output

    # If the output does not match the expected, the test failed
    diff -u $SOURCE_FILE.expected $SOURCE_FILE.output || test_diff_failed
    echo PASSED
}



test_convert_blocking()
{
    SOURCE_FILE=$1
    FLAGS=$2

    CIL_FILE=`echo -n $SOURCE_FILE| sed -e 's/.c$/.cil.c/' `

    # Preprocess
    $BINARY_CONVERT_BLOCKING "$FLAGS" $SOURCE_FILE
    mv $CIL_FILE $SOURCE_FILE.output

    cat $SOURCE_FILE.output

    # If the output does not match the expected, the test failed
    diff -u $SOURCE_FILE.expected $SOURCE_FILE.output || test_diff_failed
    echo PASSED
}



test_stack_reconstruction()
{
    SOURCE_FILE=$1
    FLAGS=$2

    CIL_FILE=`echo -n $SOURCE_FILE| sed -e 's/.c$/.cil.c/' `

    # Preprocess
    $BINARY_STACK_RECONSTRUCTION "$FLAGS" --update-version=1 --base-version $SOURCE_FILE
    mv $CIL_FILE $SOURCE_FILE.output

    cat $SOURCE_FILE.output

    # If the output does not match the expected, the test failed
    diff -u $SOURCE_FILE.expected $SOURCE_FILE.output || test_diff_failed
    echo PASSED
}



test_diff_gen()
{
    SOURCE_FILE=$1
    FLAGS=$2

    $BINARY_FULL --update-version=1 --base-version "$FLAGS" -g $SOURCE_FILE && ./a.out | grep -iv 'HCU' > $SOURCE_FILE.output 2>&1

    cat $SOURCE_FILE.output

    # If the output does not match the expected, the test failed
    diff -u $SOURCE_FILE.expected $SOURCE_FILE.output || test_diff_failed
    echo PASSED
}



test_update()
{
    SOURCE_FILE=$1
    UPDATE_FILE=$2
    FLAGS=$3
    
    UPDATE_OBJECT_FILE=`echo -n $UPDATE_FILE|sed -e 's/.c$/.o/'`
    UPDATE_SHARED_OBJECT_FILE=`echo -n $UPDATE_FILE|sed -e 's/.c$/.so/'`

    UPDATE_OBJECT_FILE_VERSION=2

    # Apply stack-reconstruction to the newer version
    $BINARY_FULL $FLAGS -g -c -fPIC $UPDATE_FILE --update-version=2 -o $UPDATE_OBJECT_FILE || test_diff_failed

    # Create a shared object from the newer version
    $BINARY_CC $FLAGS -g -shared -Wl,-soname=$UPDATE_SHARED_OBJECT_FILE.$UPDATE_OBJECT_FILE_VERSION -o lib$UPDATE_SHARED_OBJECT_FILE.$UPDATE_OBJECT_FILE_VERSION $UPDATE_OBJECT_FILE || test_diff_failed
    

    # Determine if the test case contains a static update point. If
    # so, there will be no need to start an update thread. So set
    # command-line arguments to the preprocessor that won't start the
    # thread.
    #
    # If there's a static request for an update that is not
    # commented out.
    grep '^ *HCU_STATIC_REQUEST_UPDATE' $SOURCE_FILE > /dev/null
    RETVAL=$?
    if [ $RETVAL = 0 ]; then
	STATIC_UPDATE_POINT_EXISTS=1
	STACK_RECONSTRUCTION_EXTRA_FLAGS="--no-update-thread"
    else
	STATIC_UPDATE_POINT_EXISTS=0
    fi

    # Apply stack-reconstruction to the original version
    $BINARY_FULL $STACK_RECONSTRUCTION_EXTRA_FLAGS --update-version=1 --base-version $FLAGS -g -Wl,--export-dynamic $SOURCE_FILE || test_diff_failed

    # Indent all files created to assist us in debugging
    PREPROCESSED_SOURCE_FILE=`echo -n $SOURCE_FILE.hcucc.c | sed -e 's/.c$/.cil.i/'`
    PREPROCESSED_UPDATE_FILE=`echo -n $UPDATE_FILE.hcucc.c | sed -e 's/.c$/.cil.i/'`
    indent $PREPROCESSED_SOURCE_FILE
    indent $PREPROCESSED_UPDATE_FILE

    if [ $STATIC_UPDATE_POINT_EXISTS = 1 ]; then
	# Start the original version without putting it in the background
	./a.out | grep -vi 'HCU' > $SOURCE_FILE.output 2>&1
    else
        # Start the original version
	./a.out | grep -vi 'HCU' > $SOURCE_FILE.output 2>&1 &

        # Wait for a while for the process to start, and the updating
        # thread to open its socket
	sleep 1

	# Apply the update
	$BINARY_APPLY --file=`pwd`/lib$UPDATE_SHARED_OBJECT_FILE.$UPDATE_OBJECT_FILE_VERSION --update-version=2
    fi

    cat $SOURCE_FILE.output

    # If the output does not match the expected, the test failed
    diff -u $SOURCE_FILE.expected $SOURCE_FILE.output || test_diff_failed
    echo PASSED
}



test_source_evolution()
{
    SOURCE_FILE1=$1
    SOURCE_FILE2=$2

    OUTPUT=${SOURCE_FILE1}.output
    FULL_OUTPUT=${SOURCE_FILE1}.output.full
    rm $OUTPUT
    rm $FULL_OUTPUT

    $BINARY_DIFF $SOURCE_FILE1 $SOURCE_FILE2 yes > $FULL_OUTPUT 2>&1

    EXPECTED=${SOURCE_FILE1}.expected
    cat $FULL_OUTPUT

    # We need to trim the output because the definitions that
    # remained the same are coming from included header files and
    # those may vary between machines, causing the tests to fail
    cat $FULL_OUTPUT| sed -e '1,/Computing differences/d' | sed -e '/Timings:/,$d' > $OUTPUT

    # If the output does not match the expected, the test failed
    diff -u $EXPECTED $OUTPUT || test_diff_failed
    echo PASSED
}



test_patch_gen()
{
    SOURCE_FILE1=$1
    VERSION_FILE1=$2
    SOURCE_FILE2=$3
    VERSION_FILE2=$4
    MAPPINGS_FILE=$5
    BASE_VERSION=$6

    DIFF_OUTPUT=${SOURCE_FILE1}_v${VERSION_FILE1}_to_v${VERSION_FILE2}.hcupatch.c
    OUTPUT=${SOURCE_FILE1}_v${VERSION_FILE1}_to_v${VERSION_FILE2}.hcupatch.c.output
    FULL_OUTPUT=${SOURCE_FILE1}_v${VERSION_FILE1}_to_v${VERSION_FILE2}.hcupatch.c.output.full
    UPDATE_OBJECT_FILE=`echo -n $DIFF_OUTPUT|sed -e 's/.c$/.o/'`
    rm $UPDATE_OBJECT_FILE
    rm $DIFF_OUTPUT
    rm $OUTPUT
    rm $FULL_OUTPUT

    $BINARY_BUILD_PATCH $SOURCE_FILE1 $VERSION_FILE1 $SOURCE_FILE2 $VERSION_FILE2 $MAPPINGS_FILE $BASE_VERSION yes yes

    cp $DIFF_OUTPUT $FULL_OUTPUT

    EXPECTED=$DIFF_OUTPUT.expected
    cat $FULL_OUTPUT

    # We need to trim the output because the definitions that
    # remained the same are coming from included header files and
    # those may vary between machines, causing the tests to fail
    cat $FULL_OUTPUT| sed -e '1,/Below lie the differences/d' > $OUTPUT

    # If the output does not match the expected, the test failed
    diff -u $EXPECTED $OUTPUT || test_diff_failed

    # Also, if the output does not compile, the test failed
    $BINARY_FULL $FLAGS -g -c $DIFF_OUTPUT --update-version=${VERSION_FILE2} -o $UPDATE_OBJECT_FILE || test_diff_failed
    if [ ! -e $UPDATE_OBJECT_FILE ]; then
	test_diff_failed
    fi

    echo PASSED
}



test_full_update()
{
    SOURCE_FILE1=$1
    VERSION_FILE1=$2
    SOURCE_FILE2=$3
    VERSION_FILE2=$4
    MAPPINGS_FILE=$5
    BASE_VERSION=$6
    
    # Produce the dynamic update file
    DIFF_OUTPUT=${SOURCE_FILE1}_v${VERSION_FILE1}_to_v${VERSION_FILE2}.hcupatch.c
    $BINARY_BUILD_PATCH $SOURCE_FILE1 $VERSION_FILE1 $SOURCE_FILE2 $VERSION_FILE2 $MAPPINGS_FILE $BASE_VERSION yes yes

    UPDATE_OBJECT_FILE=`echo -n $DIFF_OUTPUT|sed -e 's/.c$/.o/'`
    UPDATE_SHARED_OBJECT_FILE=`echo -n $DIFF_OUTPUT|sed -e 's/.c$/.so/'`

    # Apply stack-reconstruction to the newer version
    $BINARY_FULL -g -c -fPIC $DIFF_OUTPUT --update-version=2 -o $UPDATE_OBJECT_FILE || test_diff_failed

    # Create a shared object from the newer version
    $BINARY_CC -g -shared -Wl,-soname=$UPDATE_SHARED_OBJECT_FILE.$VERSION_FILE2 -o lib$UPDATE_SHARED_OBJECT_FILE.$VERSION_FILE2 $UPDATE_OBJECT_FILE || test_diff_failed
    

    # Determine if the test case contains a static update point. If
    # so, there will be no need to start an update thread. So set
    # command-line arguments to the preprocessor that won't start the
    # thread.
    #
    # If there's a static request for an update that is not
    # commented out.
    grep '^ *HCU_STATIC_REQUEST_UPDATE' $SOURCE_FILE1 > /dev/null
    RETVAL=$?
    if [ $RETVAL = 0 ]; then
        # This was a static request for an update
        STATIC_UPDATE_POINT_EXISTS=1

        # Also check if forking was involved
        grep 'fork *(' $SOURCE_FILE1 > /dev/null
        if [ $RETVAL = 0 ]; then
            # This test involves forking. We MUST start an update thread
            FORK_INVOLVED=1
        else
            FORK_INVOLVED=0
            STACK_RECONSTRUCTION_EXTRA_FLAGS="--no-update-thread"
        fi
    else
	STATIC_UPDATE_POINT_EXISTS=0
        FORK_INVOLVED=0
    fi

    # Apply stack-reconstruction to the original version
    $BINARY_FULL $STACK_RECONSTRUCTION_EXTRA_FLAGS --update-version=1 --base-version -g -Wl,--export-dynamic $SOURCE_FILE1 || test_diff_failed

    # Indent all files created to assist us in debugging
    PREPROCESSED_SOURCE_FILE1=`echo -n $SOURCE_FILE1.hcucc.c | sed -e 's/.c$/.cil.i/'`
    PREPROCESSED_UPDATE_FILE=`echo -n $DIFF_OUTPUT.hcucc.c | sed -e 's/.c$/.cil.i/'`
    indent $PREPROCESSED_SOURCE_FILE1
    indent $PREPROCESSED_UPDATE_FILE

    if [ $STATIC_UPDATE_POINT_EXISTS = 1 ]; then
	# Start the original version without putting it in the background
	./a.out > $SOURCE_FILE1.temp.output 2>&1
    else
        # Start the original version
        ./a.out > $SOURCE_FILE1.temp.output 2>&1 &

        # Wait for a while for the process to start, and the updating
        # thread to open its socket
	sleep 2

	# Apply the update
	$BINARY_APPLY --file=`pwd`/lib$UPDATE_SHARED_OBJECT_FILE.$VERSION_FILE2 --update-version=2
    fi

    if [ $STATIC_UPDATE_POINT_EXISTS = 0 ]; then
        # A dynamic update was applied. Wait for the output to be
        # produced
        sleep 3
    elif [ $FORK_INVOLVED = 1 ]; then
        # A static update that involved a fork was applied. Wait for
        # the output to be produced
	sleep 3
    fi
    cat $SOURCE_FILE1.temp.output | grep -vi 'HCU' > $SOURCE_FILE1.output
    cat $SOURCE_FILE1.output

    # If the output does not match the expected, the test failed
    diff -u $SOURCE_FILE1.expected $SOURCE_FILE1.output || test_diff_failed
    echo PASSED
}



test_diff_failed()
{
    echo FAILED
    exit 0
}
